#include "d3d.h"
#include "d3drm.h"


// atgpRender - gives control over rendering in an RM app (DX5 Interfaces)
// see http://www.atgp.com/rblts/coding.htm#atgpRender for details
// (C) 1999 Above the Garage Productions
// All Rights Reserved - except you are free to use this in your own code provided you don't hold me liable for anything
// and you give me (Stephen Clarke-Willson d.b.a. Above the Garage Productions) credit for inspiring this.

bool atgpIsaMesh(IDirect3DRMObject* obj)
{
	IDirect3DRMMeshBuilder2* testMesh;
	try {
		if (obj && SUCCEEDED(obj->QueryInterface(IID_IDirect3DRMMeshBuilder, (void **) &testMesh))) {
			testMesh->Release();
			return TRUE;
		}
	}
	catch (...) {
		OutputDebugString("atgpIsaMesh exception\n");
		return FALSE;
	}
	return FALSE;
}

void atgpRender(IDirect3DRMDevice2* dev, IDirect3DRMViewport* vp, IDirect3DRMFrame2* scene, IDirect3DRMFrame2* camera, int depth)
{
	HRESULT hr = 0;
	// This is an optimization to reuse the memory as much as possible via a high-water mark
	static IDirect3D2* d3d = NULL;
	static IDirect3DDevice2*  d3ddev = NULL;
	static IDirect3DViewport* d3dvp = NULL;
	static D3DVECTOR* vertices = NULL;
	static D3DVECTOR* normals = NULL;
	static D3DLVERTEX* d3dverts = NULL;
	static unsigned short* fdata = NULL;
	static D3DMATERIAL material;
	static D3DMATERIALHANDLE material_handle;
	static IDirect3DMaterial2* d3dmat = NULL;
	static int fmax = 0;
	static unsigned long* face_data = NULL;
	static unsigned long vertmax = 0;
	static unsigned long normmax = 0;
	static unsigned long facemax = 0;
	D3DCOLOR white = RGB(1.0f, 1.0f, 1.0f);
	D3DCOLOR bronze = RGB(1.0f, 0.8f, 0.5f);

	if (depth == 0) {
		// Set up D3D device
		hr = dev->GetDirect3DDevice2(&d3ddev);
		// Get D3D interface
		hr = d3ddev->GetDirect3D(&d3d);
		
		// Set up the material structure, com object, and handle (lame, huh?)
		hr = d3d->CreateMaterial(&d3dmat, 0);
		ZeroMemory(&material, sizeof(D3DMATERIAL)); 
		material.dwSize   = sizeof(D3DMATERIAL);
 
		// Set the RGBA for diffuse reflection.
		material.dcvDiffuse.r = (D3DVALUE)1.0; 
		material.dcvDiffuse.g = (D3DVALUE)0.8; 
		material.dcvDiffuse.b = (D3DVALUE)0.5; 
		material.dcvDiffuse.a = (D3DVALUE)1.0; 
 
		// Set the RGBA for ambient reflection. 
		material.dcvAmbient.r = (D3DVALUE)0.3; 
		material.dcvAmbient.g = (D3DVALUE)0.3; 
		material.dcvAmbient.b = (D3DVALUE)0.3; 
		material.dcvAmbient.a = (D3DVALUE)1.0; 
 
		// Set the color and sharpness of specular highlights. 
		material.dcvSpecular.r = (D3DVALUE)1.0; 
		material.dcvSpecular.g = (D3DVALUE)1.0; 
		material.dcvSpecular.b = (D3DVALUE)1.0; 
		material.dcvSpecular.a = (D3DVALUE)1.0; 
		material.dvPower = (float)50.0; 
 
		// Use a 16 entry color ramp, in case 
		// we're using ramp emulation. 
		material.dwRampSize = 16; 
		hr = d3dmat->SetMaterial(&material);
		hr = d3dmat->GetHandle(d3ddev, &material_handle);
		
		// Set up viewport
		hr = vp->GetDirect3DViewport(&d3dvp);
	}

	// For each child of f, call atgpRender
	// This makes it depth first
	IDirect3DRMFrameArray* fa;
	scene->GetChildren(&fa);
	unsigned nchildren = fa->GetSize();
//	TRACE("atgpRender(%d) nchildren %d\n", depth, nchildren);
	unsigned i, j;
	for (i = 0; i < nchildren; i++) {
		IDirect3DRMFrame2* child;
		fa->GetElement(i, (IDirect3DRMFrame**)&child);
		atgpRender(dev, vp, child, camera, depth + 1);
		child->Release();
	}
	// get here when no more children - must only be visuals or nothing left
	IDirect3DRMVisualArray* va;
	scene->GetVisuals(&va);
	unsigned nvisuals = va->GetSize();
//	TRACE("atgpRender(%d): nvisuals %d\n", depth, nvisuals);
	for(i = 0; i < nvisuals; i++) {
		// get each visual - only hang onto meshes
		IDirect3DRMVisual* vis;
		va->GetElement(i, &vis);
		// test type of visual
		if (atgpIsaMesh(vis)) {
			// retrieve faces
			IDirect3DRMMeshBuilder2* m = (IDirect3DRMMeshBuilder2*)vis;
			unsigned long vertcount = m->GetVertexCount();
			unsigned long normcount = vertcount; 
			unsigned long facecount = 0;
			if (vertcount > vertmax) {
				if (vertices)
					free(vertices);
				vertices = (D3DVECTOR*)malloc(vertcount * sizeof(D3DVECTOR));
				vertmax = vertcount;
				if (d3dverts)
					free(d3dverts);
				d3dverts = (D3DLVERTEX*)malloc(vertcount * sizeof(D3DLVERTEX));
			}
			if (normcount > normmax) {
				if (normals)
					free(normals);
				normals = (D3DVECTOR*)malloc(vertcount * sizeof(D3DVECTOR));
				normmax = normcount;
			}
			m->GetVertices(&vertcount, vertices, &normcount, normals, &facecount, NULL);
			if (facecount > facemax) {
				if (face_data)
					free(face_data);
				face_data = (unsigned long*)malloc(facecount * sizeof(int));
				facemax = facecount;
			}
			m->GetVertices(&vertcount, vertices, &normcount, normals, &facecount, face_data);
			// Transform everything to world space and pack into D3DVERTEX array
			for (j = 0; j < vertcount; j++) {
				D3DVECTOR  newVector, newNormal;
				scene->Transform(&newVector, &vertices[j]);
				scene->Transform(&newNormal, &normals[j]);
				D3DRMVectorNormalize(&newNormal);
				D3DLVERTEX& vert = d3dverts[j];
				vert.x = newVector.x;
				vert.y = newVector.y;
				vert.z = -newVector.z;
//				vert.nx = newNormal.x;
//				vert.ny = newNormal.y;
//				vert.nz = newNormal.z;
				vert.tu = 0.0f;
				vert.tv = 0.0f;
				vert.dcColor = bronze;
				vert.dcSpecular = bronze;
			}
			int index = 0;
			while (face_data[index] != 0) {
				int fcount = face_data[index++];
				if (fcount > fmax) {
					if (fdata)
						free(fdata);
					fdata = (unsigned short*)malloc(fcount * sizeof(unsigned short));
					fmax = fcount;
				}
				int k;
				for(k = 0; k < fcount; k++) {
					fdata[k] = (unsigned short)face_data[index++];
					if (normcount)
						index++;
				}
				// Now call DrawPrimitive on this piece (fan) of the mesh
				D3DMATRIX newWorldMatrix = {1.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 1.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 1.0f, 0.0f,
											0.0f, 0.0f, 0.0f, 1.0f};
				D3DMATRIX newViewMatrix = {	1.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 1.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 1.0f, 0.0f,
											0.0f, 0.0f, 25.0f, 1.0f};
				D3DMATRIX oldWorldMatrix;
				D3DMATRIX oldViewMatrix;
				hr = d3ddev->GetTransform(D3DTRANSFORMSTATE_WORLD, &oldWorldMatrix);
				hr = d3ddev->SetTransform(D3DTRANSFORMSTATE_WORLD, &newWorldMatrix);
				hr = d3ddev->GetTransform(D3DTRANSFORMSTATE_VIEW, &oldViewMatrix);
				hr = d3ddev->SetTransform(D3DTRANSFORMSTATE_VIEW, &newViewMatrix);
				hr = d3ddev->SetRenderState(D3DRENDERSTATE_TEXTUREHANDLE, NULL);
				hr = d3ddev->SetLightState(D3DLIGHTSTATE_AMBIENT, bronze);			
				hr = d3ddev->SetLightState(D3DLIGHTSTATE_MATERIAL, NULL);
			//	hr = d3ddev->SetLightState(D3DLIGHTSTATE_MATERIAL, material_handle);	// hack for HUD - should look up color from Meshbuilder->Face
				hr = d3ddev->SetRenderState(D3DRENDERSTATE_FILLMODE, D3DFILL_SOLID);
				hr = d3ddev->SetRenderState(D3DRENDERSTATE_SHADEMODE, D3DSHADE_GOURAUD);				
				hr = d3ddev->DrawIndexedPrimitive(D3DPT_TRIANGLEFAN, D3DVT_VERTEX, d3dverts, vertcount, fdata, fcount, D3DDP_WAIT);
				hr = d3ddev->SetTransform(D3DTRANSFORMSTATE_WORLD, &oldWorldMatrix);
				hr = d3ddev->SetTransform(D3DTRANSFORMSTATE_VIEW, &oldViewMatrix);
			}
#if 0
			// memory is not freed until app exits
			free(vertices);
			free(normals);
			free(face_data);
#endif
		}
		vis->Release();
	}
	va->Release();
	fa->Release();
	if (depth == 0) {
		d3dvp->Release();
		d3dvp = NULL;
		d3dmat->Release();
		d3dmat = NULL;
		d3d->Release();
		d3d = NULL;
		d3ddev->Release();
		d3ddev = NULL;
	}
}

